import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

def run_lwlr(file_path, tau=0.5):
    data = pd.read_csv(file_path)

    X_train = np.array(data['total_bill'])[:, np.newaxis]
    y_train = np.array(data['tip'])

    print("Training input length:", len(X_train))

    X_test = np.array([i / 10 for i in range(500)])[:, np.newaxis]
    y_test = []

    for r in range(len(X_test)):
        # Compute weights
        wts = np.exp(-np.sum((X_train - X_test[r]) ** 2, axis=1) / (2 * tau ** 2))
        W = np.diag(wts)

        # Compute parameters theta
        try:
            theta = np.linalg.inv(X_train.T @ W @ X_train) @ X_train.T @ W @ y_train
            prediction = X_test[r] @ theta
            y_test.append(prediction)
        except np.linalg.LinAlgError:
            y_test.append(0)  # Fallback if matrix is non-invertible

    print("Prediction count:", len(y_test))

    # Plotting results
    plt.figure(figsize=(10, 6))
    plt.plot(X_train.squeeze(), y_train, 'o', label="Training Data")
    plt.plot(X_test.squeeze(), y_test, '-', label="LWLR Prediction")
    plt.xlabel("Total Bill")
    plt.ylabel("Tip")
    plt.title("Locally Weighted Linear Regression")
    plt.legend()
    plt.grid(True)
    plt.show()



import csv

def candidate_elimination(file_path):
    with open(file_path, 'r') as f:
        reader = csv.reader(f)
        data = list(reader)[1:]  # skip header

    concepts = [row[:-1] for row in data]
    targets = [row[-1] for row in data]

    n = len(concepts[0])
    S = concepts[0].copy()
    G = [['?' for _ in range(n)]]

    for i, example in enumerate(concepts):
        if targets[i] == "Yes":
            for j in range(n):
                if S[j] != example[j]:
                    S[j] = '?'
            G = [g for g in G if all(g[k] == '?' or g[k] == S[k] for k in range(n))]
        else:
            new_G = []
            for g in G:
                for j in range(n):
                    if g[j] == '?':
                        if S[j] != '?':
                            g_new = g.copy()
                            g_new[j] = S[j]
                            new_G.append(g_new)
            G = new_G

    print("Final Specific Hypothesis (S):", S)
    print("Final General Hypotheses (G):", G)



import numpy as np
from collections import Counter
from sklearn.tree import DecisionTreeClassifier
from sklearn import datasets
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score, confusion_matrix

class RandomForest:
    def __init__(self, n_trees=10, max_depth=None):
        self.n_trees = n_trees
        self.max_depth = max_depth
        self.trees = []

    def fit(self, X, y):
        self.trees = []
        for _ in range(self.n_trees):
            tree = DecisionTreeClassifier(max_depth=self.max_depth)
            idxs = np.random.choice(len(X), len(X), replace=True)
            X_sample, y_sample = X[idxs], y[idxs]
            tree.fit(X_sample, y_sample)
            self.trees.append(tree)

    def predict(self, X):
        predictions = np.array([tree.predict(X) for tree in self.trees])
        majority_votes = np.swapaxes(predictions, 0, 1)
        return np.array([Counter(row).most_common(1)[0][0] for row in majority_votes])

if __name__ == "__main__":
    # Load dataset
    X, y = datasets.load_breast_cancer(return_X_y=True)

    # Split dataset
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=1)

    # Train Random Forest
    model = RandomForest(n_trees=10, max_depth=10)
    model.fit(X_train, y_train)

    # Predict and evaluate
    y_pred = model.predict(X_test)

    print("Accuracy:", accuracy_score(y_test, y_pred))
    print("Confusion Matrix:\n", confusion_matrix(y_test, y_pred))



import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from sklearn.mixture import GaussianMixture
from sklearn.cluster import KMeans

def run_clustering(file_path, k=3, gmm_components=3):
    data = pd.read_csv(file_path)
    print("Input Data Shape:", data.shape)
    print(data.head())

    f1 = data['V1'].values
    f2 = data['V2'].values
    X = np.array(list(zip(f1, f2)))

    print("\nOriginal Data Points (X):\n", X)

    # Plot original dataset
    print('\nGraph for Whole Dataset')
    plt.scatter(f1, f2, c='black', s=7)
    plt.title("Original Dataset")
    plt.xlabel("V1")
    plt.ylabel("V2")
    plt.show()

    # KMeans Clustering
    kmeans = KMeans(n_clusters=k, random_state=0)
    kmeans_labels = kmeans.fit_predict(X)
    centroids = kmeans.cluster_centers_

    print("\nKMeans Cluster Labels:\n", kmeans_labels)
    print("KMeans Centroids:\n", centroids)

    print('\nGraph using KMeans Algorithm')
    plt.scatter(X[:, 0], X[:, 1], c=kmeans_labels, s=40, cmap='viridis')
    plt.scatter(centroids[:, 0], centroids[:, 1], marker='*', s=200, c='red')
    plt.title("KMeans Clustering")
    plt.xlabel("V1")
    plt.ylabel("V2")
    plt.show()

    # EM using GMM
    gmm = GaussianMixture(n_components=gmm_components, random_state=0).fit(X)
    gmm_labels = gmm.predict(X)
    probs = gmm.predict_proba(X)
    size = 10 * probs.max(1) ** 3

    print('\nGraph using EM Algorithm (GMM)')
    plt.scatter(X[:, 0], X[:, 1], c=gmm_labels, s=size, cmap='viridis')
    plt.title("EM Clustering using GMM")
    plt.xlabel("V1")
    plt.ylabel("V2")
    plt.show()



import pandas as pd
import math
from collections import Counter

# Calculate entropy of a dataset
def entropy(data):
    label_column = data.iloc[:, -1]
    counts = Counter(label_column)
    total = len(label_column)
    return -sum((count / total) * math.log2(count / total) for count in counts.values())

# Information gain of splitting on an attribute
def info_gain(data, attr):
    total_entropy = entropy(data)
    values = data[attr].unique()
    weighted_entropy = 0
    for val in values:
        subset = data[data[attr] == val]
        weighted_entropy += (len(subset) / len(data)) * entropy(subset)
    return total_entropy - weighted_entropy

# ID3 algorithm
def id3(data, features):
    labels = data.iloc[:, -1]
    
    # If all labels are same
    if len(labels.unique()) == 1:
        return labels.iloc[0]

    # If no more features left
    if not features:
        return labels.mode()[0]

    # Choose best attribute
    gains = [info_gain(data, attr) for attr in features]
    best_attr = features[gains.index(max(gains))]

    tree = {best_attr: {}}
    for val in data[best_attr].unique():
        subset = data[data[best_attr] == val]
        subtree = id3(subset.drop(columns=[best_attr]), [f for f in features if f != best_attr])
        tree[best_attr][val] = subtree

    return tree

# Classify a test sample
def classify(tree, sample):
    if not isinstance(tree, dict):
        return tree
    root = next(iter(tree))
    value = sample.get(root)
    if value in tree[root]:
        return classify(tree[root][value], sample)
    else:
        return "Unknown"

# Main function to run ID3 and classify a sample
def decision_tree_from_csv(file_path):
    data = pd.read_csv(file_path)
    features = list(data.columns[:-1])
    tree = id3(data, features)

    print("\nGenerated Decision Tree:")
    print(tree)

    # Classify one sample (hardcoded, as required)
    test_sample = {
        features[i]: data.iloc[0][i] for i in range(len(features))
    }
    print("\nClassifying sample:", test_sample)
    result = classify(tree, test_sample)
    print("Predicted label:", result)



import numpy as np
import matplotlib.pyplot as plt
from sklearn.model_selection import train_test_split
from sklearn import datasets

class SVM:
    def __init__(self, learning_rate=0.001, lambda_param=0.01, n_iters=1000):
        self.lr = learning_rate
        self.lambda_param = lambda_param
        self.n_iters = n_iters
        self.w = None
        self.b = None

    def fit(self, X, y):
        n_samples, n_features = X.shape
        y_ = np.where(y <= 0, -1, 1)
        self.w = np.zeros(n_features)
        self.b = 0

        for _ in range(self.n_iters):
            for idx, x_i in enumerate(X):
                condition = y_[idx] * (np.dot(x_i, self.w) - self.b) >= 1
                if condition:
                    self.w -= self.lr * (2 * self.lambda_param * self.w)
                else:
                    self.w -= self.lr * (2 * self.lambda_param * self.w - np.dot(x_i, y_[idx]))
                    self.b -= self.lr * y_[idx]

    def predict(self, X):
        approx = np.dot(X, self.w) - self.b
        return np.sign(approx)

def accuracy(y_true, y_pred):
    return np.sum(y_true == y_pred) / len(y_true)

def run_svm_demo():
    # Generate synthetic data
    X, y = datasets.make_blobs(n_samples=50, n_features=2, centers=2, cluster_std=1.05, random_state=40)
    y = np.where(y == 0, -1, 1)

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=123)

    clf = SVM()
    clf.fit(X_train, y_train)
    predictions = clf.predict(X_test)

    print("SVM Classification Accuracy:", accuracy(y_test, predictions))

    # Visualization
    def get_hyperplane_value(x, w, b, offset):
        return (-w[0] * x + b + offset) / w[1]

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    plt.scatter(X[:, 0], X[:, 1], marker="o", c=y)

    x0_1 = np.amin(X[:, 0])
    x0_2 = np.amax(X[:, 0])

    x1_1 = get_hyperplane_value(x0_1, clf.w, clf.b, 0)
    x1_2 = get_hyperplane_value(x0_2, clf.w, clf.b, 0)

    x1_1_m = get_hyperplane_value(x0_1, clf.w, clf.b, -1)
    x1_2_m = get_hyperplane_value(x0_2, clf.w, clf.b, -1)

    x1_1_p = get_hyperplane_value(x0_1, clf.w, clf.b, 1)
    x1_2_p = get_hyperplane_value(x0_2, clf.w, clf.b, 1)

    ax.plot([x0_1, x0_2], [x1_1, x1_2], "y--")
    ax.plot([x0_1, x0_2], [x1_1_m, x1_2_m], "k")
    ax.plot([x0_1, x0_2], [x1_1_p, x1_2_p], "k")

    x1_min = np.amin(X[:, 1])
    x1_max = np.amax(X[:, 1])
    ax.set_ylim([x1_min - 3, x1_max + 3])

    plt.title("Custom SVM Decision Boundary")
    plt.show()



import numpy as np
from sklearn.datasets import load_iris
from sklearn.model_selection import train_test_split
from sklearn.neighbors import KNeighborsClassifier
from sklearn import metrics

def run_knn_iris(test_size=0.4, random_state=1, neighbors=1):
    iris = load_iris()
    x = iris.data
    y = iris.target

    print("Sample input data:\n", x[:5])
    print("Sample target values:\n", y[:5])

    # Split the dataset
    xtrain, xtest, ytrain, ytest = train_test_split(x, y, test_size=test_size, random_state=random_state)

    print("\nTotal Dataset Shape:", iris.data.shape)
    print("Training set size:", len(xtrain))
    print("Test set size:", len(ytest))

    # Train KNN
    knn = KNeighborsClassifier(n_neighbors=neighbors)
    knn.fit(xtrain, ytrain)

    # Predict
    pred = knn.predict(xtest)

    # Accuracy
    accuracy = metrics.accuracy_score(ytest, pred)
    print("\nAccuracy:", accuracy * 100, "%")

    # Display predictions in label form
    ytestn = [iris.target_names[i] for i in ytest]
    predn = [iris.target_names[i] for i in pred]

    print("\n  Predicted    Actual")
    for i in range(len(pred)):
        print(f"{i:<3} {predn[i]:<10} {ytestn[i]:<10}")



import csv
import math
import random
import statistics

def calculate_probability(x, mean, stdev):
    exponent = math.exp(-(math.pow(x - mean, 2) / (2 * math.pow(stdev, 2))))
    return (1 / (math.sqrt(2 * math.pi) * stdev)) * exponent

def run_naive_bayes(file_path):
    dataset = []
    
    with open(file_path) as csvfile:
        reader = csv.reader(csvfile)
        for row in reader:
            dataset.append([float(attr) for attr in row])

    dataset_size = len(dataset)
    print('Dataset size:', dataset_size)

    train_size = int(0.7 * dataset_size)

    X_train = []
    X_test = dataset.copy()
    training_indexes = random.sample(range(dataset_size), train_size)

    for i in training_indexes:
        X_train.append(dataset[i])
        X_test.remove(dataset[i])

    # Separate training data by class
    classes = {}
    for sample in X_train:
        label = int(sample[-1])
        if label not in classes:
            classes[label] = []
        classes[label].append(sample)

    # Calculate mean and stdev for each class
    summaries = {}
    for class_value, samples in classes.items():
        summary = [(statistics.mean(attr), statistics.stdev(attr)) for attr in zip(*samples)]
        del summary[-1]  # Remove class column stats
        summaries[class_value] = summary

    # Predict labels for test data
    predictions = []
    for sample in X_test:
        probabilities = {}
        for class_value, class_summary in summaries.items():
            probabilities[class_value] = 1
            for i, attr in enumerate(class_summary):
                probabilities[class_value] *= calculate_probability(sample[i], attr[0], attr[1])
        
        predicted_label = max(probabilities, key=probabilities.get)
        predictions.append(predicted_label)

    # Calculate accuracy
    correct = 0
    for i in range(len(X_test)):
        if X_test[i][-1] == predictions[i]:
            correct += 1

    accuracy = (correct / len(X_test)) * 100
    print(f"Accuracy: {accuracy:.2f}%")
